//*************************************************************************
//* ====================
//*  THelicalTrack Class
//* ====================
//*
//* (Description)
//*   A class to implement a helical track object.
//* (Requires)
//*     TVTrack, TCylinder, TCircle, TVector3, TMatrixD
//* (Provides)
//*     class THelicalTrack
//* (Update Recored)
//*   2003/10/03  K.Fujii       Original version.
//*
//*************************************************************************
//

#include <iostream>
#include "THelicalTrack.h"

using namespace std;

//_____________________________________________________________________
//  -----------------------------------
//  Helical Track Class
//  -----------------------------------
//
//_____________________________________________________________________
//  --------------
//  Ctors and Dtor
//  --------------
//

THelicalTrack::THelicalTrack(Double_t dr,
                             Double_t phi0,
                             Double_t kappa,
                             Double_t dz,
                             Double_t tanl,
                             Double_t x0,
                             Double_t y0,
                             Double_t z0,
                             Double_t b)
             : TVTrack(dr,phi0,kappa,dz,tanl, x0,y0,z0, b)
{
}

THelicalTrack::THelicalTrack(const TMatrixD &a, const TVector3 &x0, Double_t b)
             : TVTrack(a, x0, b)
{
}

//_____________________________________________________________________
//  ----------------
//  Utility methods
//  ----------------

void THelicalTrack::MoveTo(const TVector3 &xv0to, // new pivoit
                                 Double_t &fid,   // deflection angle
                                 TMatrixD *FPtr,  // propagator matrix
                                 TMatrixD *CPtr)  // covariance matrix
{
   // ---------------------------------------------------
   // (0) Preparation
   // ---------------------------------------------------
   //   Define some numerical constants.
   //

   static const Double_t kPi    = TMath::Pi();
   static const Double_t kTwoPi = 2.0*kPi;

   //   Copy helix parmeters to local variables

   Double_t dr    = fDrho;
   Double_t fi0   = fPhi0;
   while (fi0 < 0.)      fi0 += kTwoPi;
   while (fi0 > kTwoPi)  fi0 -= kTwoPi;
   Double_t cpa   = fKappa;
   Double_t dz    = fDz;
   Double_t tnl   = fTanL;

   Double_t x0    = fX0.X();
   Double_t y0    = fX0.Y();
   Double_t z0    = fX0.Z();
   Double_t xv    = xv0to.X();
   Double_t yv    = xv0to.Y();
   Double_t zv    = xv0to.Z();

   // ---------------------------------------------------
   // (1) Calculate a' = f_k-1(a_k-1)
   // ---------------------------------------------------
   //        a' = (dr', fi0', cpa', dz', tnl')
   //        a  = (dr , fi0 , cpa , dz , tnl )
   //

   Double_t r     = fAlpha/cpa;
   Double_t rdr   = r + dr;
   Double_t csf0  = TMath::Cos(fi0);
   Double_t snf0  = TMath::Sqrt(TMath::Max(0.0, (1.0-csf0)*(1.0+csf0)));
   if (fi0 > kPi) snf0 = -snf0;

   Double_t xc    = x0 + rdr*csf0;
   Double_t yc    = y0 + rdr*snf0;
   Double_t fi0p  = 0.;

   if (cpa > 0.) fi0p = TMath::ATan2((yc-yv),(xc-xv));
   if (cpa < 0.) fi0p = TMath::ATan2((yv-yc),(xv-xc));
   while (fi0p < 0.)      fi0p += kTwoPi;
   while (fi0p > kTwoPi)  fi0p -= kTwoPi;

   Double_t csf   = TMath::Cos(fi0p);
   Double_t snf   = TMath::Sqrt(TMath::Max(0.0, (1.0-csf)*(1.0+csf)));
   if (fi0p > kPi) snf = -snf;

   Double_t anrm  = 1.0/TMath::Sqrt(csf*csf+snf*snf);
            csf  *= anrm;
            snf  *= anrm;
   Double_t csfd  = csf*csf0 + snf*snf0;
   Double_t snfd  = snf*csf0 - csf*snf0;

   fid   = fi0p - fi0;
   while (fid < 0)      fid += kTwoPi;
   while (fid > kTwoPi) fid -= kTwoPi;
   if    (fid > kPi)    fid -= kTwoPi;

   Double_t drp   = (xc-xv)*csf + (yc-yv)*snf - r;
   Double_t dzp   = z0 - zv + dz - r*tnl*fid;

   TMatrixD av(5,1);
   av(0,0) = drp;
   av(1,0) = fi0p;
   av(2,0) = cpa;
   av(3,0) = dzp;
   av(4,0) = tnl;

   THelicalTrack helto(av,xv0to);
   helto.fAlpha = fAlpha;

   if (!FPtr && !CPtr) {
      *this = helto;
      return;
   }

   TMatrixD F(5,5);

   // ---------------------------------------------------
   // (2) Calculate @a'/@a = @a'/a = F_k-1
   // ---------------------------------------------------
   //        a' = (dr', fi0', cpa', dz', tnl')
   //        a  = (dr , fi0 , cpa , dz , tnl )
   //

   Double_t rdrpr = 1.0/(r+drp);
   Double_t rcpar = r/cpa;

   // @drho'/@a
   F(0,0) = csfd;
   F(0,1) = rdr*snfd;
   F(0,2) = rcpar*(1.0-csfd);
   F(0,3) = 0;
   F(0,4) = 0;

   // @phi0'/@a
   F(1,0) = -rdrpr*snfd;
   F(1,1) =  rdr*rdrpr*csfd;
   F(1,2) =  rcpar*rdrpr*snfd;
   F(1,3) =  0;
   F(1,4) =  0;

   // @kappa'/@a
   F(2,0) = 0;
   F(2,1) = 0;
   F(2,2) = 1;
   F(2,3) = 0;
   F(2,4) = 0;

   // @dz'/@a
   F(3,0) =  r*rdrpr*tnl*snfd;
   F(3,1) =  r*tnl*(1.0-rdr*rdrpr*csfd);
   F(3,2) =  rcpar*tnl*(fid-r*rdrpr*snfd);
   F(3,3) =  1;
   F(3,4) = -r*fid;

   // @tanl'/@a
   F(4,0) = 0;
   F(4,1) = 0;
   F(4,2) = 0;
   F(4,3) = 0;
   F(4,4) = 1;

   if (FPtr) *FPtr = F;
   if (!CPtr) {
      *this = helto;
      return;
   }

   // ---------------------------------------------------
   // (3) Calculate C' = C^k-1_k
   // ---------------------------------------------------

   TMatrixD &C  = *CPtr;
   TMatrixD  Ft = TMatrixD(TMatrixD::kTransposed, F);
   TMatrixD  Cp = F * C * Ft;
   C = Cp;

   *this = helto;
}

TVector3 THelicalTrack::CalcXAt(Double_t phi) const
{
   Double_t csf0 = TMath::Cos(fPhi0);
   Double_t snf0 = TMath::Sin(fPhi0);
   Double_t snfd = TMath::Sin(fPhi0 + phi);
   Double_t csfd = TMath::Cos(fPhi0 + phi);
   Double_t rho  = fAlpha/fKappa;

   Double_t x    = fX0.X() + fDrho * csf0 + rho * (csf0 - csfd);
   Double_t y    = fX0.Y() + fDrho * snf0 + rho * (snf0 - snfd);
   Double_t z    = fX0.Z() + fDz          - rho * fTanL * phi;

   return TVector3(x,y,z);
}

TMatrixD THelicalTrack::CalcDxDa(Double_t phi) const
{
   Double_t fi0   = fPhi0;
   Double_t r     = fAlpha/fKappa;
   Double_t rcpar = r/fKappa;

   Double_t snf0 = TMath::Sin(fi0);
   Double_t csf0 = TMath::Cos(fi0);
   Double_t snfd = TMath::Sin(fi0 + phi);
   Double_t csfd = TMath::Cos(fi0 + phi);

   TMatrixD dxda(3,5);
   // @x/@a
   dxda(0,0) =  csf0;
   dxda(0,1) = -fDrho * snf0 - r * (snf0 - snfd);
   dxda(0,2) = -rcpar * (csf0 - csfd);
   dxda(0,3) =  0;
   dxda(0,4) =  0;

   // @y/@a
   dxda(1,0) =  snf0;
   dxda(1,1) =  fDrho * csf0 + r *(csf0 - csfd);
   dxda(1,2) = -rcpar * (snf0 - snfd);
   dxda(1,3) =  0;
   dxda(1,4) =  0;

   // @z/@a
   dxda(2,0) = 0;
   dxda(2,1) = 0;
   dxda(2,2) = rcpar * phi * fTanL;
   dxda(2,3) = 1;
   dxda(2,4) = - r * phi;

   return dxda;
}

TMatrixD THelicalTrack::CalcDxDphi(Double_t phi) const
{
   Double_t r    = fAlpha/fKappa;

   Double_t snfd = TMath::Sin(fPhi0 + phi);
   Double_t csfd = TMath::Cos(fPhi0 + phi);

   TMatrixD dxdphi(3,1);
   dxdphi(0,0) =  r * snfd;
   dxdphi(1,0) = -r * csfd;
   dxdphi(2,0) = -r * fTanL;

   return dxdphi;
}
